searchNodes=[{"doc":"⚠️ Preview, requires Elixir 1.11.0-dev to runDomo is a library for defining custom composable types for fields of a struct to make these pieces of data to flow through the app consistently.The library aims for two goals:to model a business domain entity's possible valid states with custom types for fields of struct representing the entityto verify that entity structs are assembled to one of the allowed valid states at compile-timeIt's a library to define what piece of data is what and make a compiler and dialyzer to cover one's back, reminding about taken definitions.The validation of the incoming data is on the author of the concrete application. The library can only ensure the consistent processing of that valid data throughout the system.The library has the means to build structs and depends on the TypedStruct to do so.RationaleTo model a business domain entity, one may define a named struct with several fields of primitive data types. Construction of the struct from parsed data can look like this:%Order{ id: &quot;156&quot;, quantity: 2.5 note: &quot;Deliver on Tue&quot; }and modification of the struct's data can be done with a function of the following signature:@spec put_quantity(Order.t(), float()) :: Order.t() def put_quantity(order, quantity) ...Primitive types of binary and float are universal and have no relation to the Order struct specifically. That is, any data of these types can leak into the new struct instance by mistake. The float type defining quantity reflects no measure from the business domain. Meaning, that a new requirement - to measure quantity in Kilograms or Units makes space for misinterpretation of the quantity field's value processed in any part of the app.How about some domain modeling?In the context given above, it'd be great to define a contract to allow only valid states for Order struct fields, that enables:local reasoning about the relation of value to the struct in any nested function of the appcompile-time verification of assembling/updating of the structure from values that relates only to itOne possible valid way to do this is to use Domo library like the following:defmodule Order do use Domo deftag Id, for_type: String.t() deftag Quantity do for_type __MODULE__.Kilograms.t() | __MODULE__.Units.t() deftag Kilograms, for_type: float deftag Units, for_type: integer end deftag Note, for_type: :none | String.t() typedstruct do field :id, Id.t() field :quantity, Quantity.t() field :note, Note.t(), default: Note --- :none end endThen the construction of the struct becomes like this:Order.new!(%{ id: Id --- &quot;156&quot;, quantity: Quantity --- Kilograms --- 2.5 note: Note --- &quot;Deliver on Tue&quot; })And a signature of a custom function to modify the struct becomes like this:@spec put_quantity(Order.t(), Order.Quantity.t()) :: Order.t() def put_quantity(order, Quantity --- Units --- units) ... def put_quantity(order, Quantity --- Kilograms --- kilos) ...Thanks to the Domo library, every field of the structure becomes a tagged tuple consisting of a tag and a value. A tag is a module itself. Several tags can be nested, defining valid tag chains as a shape for values of primitive type. That makes it possible to do pattern matching against the shape of the struct's value. That enables the dialyzer to validate contracts for the structure itself and the structure's field values.UsageSetupTo use Domo in your project, add this to your Mix dependencies:{:domo, &quot;~&gt; 0.0.7&quot;}To avoid mix format putting parentheses on tagged tuples definitions made with ---/2 operator, you can add to your .formatter.exs:[ ..., import_deps: [:typed_struct] ]General usageDefine a tagTo define a tag on the top level of a file import Domo, then define the tag name and type associated value with deftag/2 macro.import Domo deftag Title, for_type: String.t() deftag Height do for_type: __MODULE__.Meters.t() | __MODULE__.Foots.t() deftag Meters, for_type: float deftag Foots, for_type: float endAny tag is a module by itself. Type t() of the tag is a tagged tuple. When defining a tag in a block form, you can specify the associated value type through the for_type/1 macro.To add a tag or a tag chain to a value use ---/2 macro.alias Height.{Meters, Foots} t = Title --- &quot;Eiffel tower&quot; m = Height --- Meters --- 324.0 f = Height --- Foots --- 1062.992Under the hood, the tag chain is a series of nested tagged tuples where the value is in the core. Because of that, you can use the ---/2 macro in pattern matching.{Height, {Meters, 324.0}} == m @spec to_string(Height.t()) :: String.t() def to_string(Height --- Meters --- val), do: to_string(val) &lt;&gt; &quot; m&quot; def to_string(Height --- Foots --- val), do: to_string(val) &lt;&gt; &quot; ft&quot;Each tag module has type t() of tagged tuple with the name of tag itself and a value type specified with for_type. Use t() in the function spec to inform the dialyzer about the tagged argument.Define a structureTo define a structure with field value's contracts, use Domo, then define your struct with a typedstruct/1 block.defmodule Order do use Domo deftag Id, for_type: String.t() deftag Note, for_type: :none | String.t() @typedoc &quot;An Order from Sales context&quot; typedstruct do field :id, Id.t() field :note, Note.t(), default: Note --- :none end endEach field is defined through field/3 macro. The generated structure has all fields enforced, default values specified by default: key, and type t() constructed with field types. See TypedStruct library documentation for implementation details.Use new!/1 and put!/3 functions that are automatically defined for the struct to create a new instance and update an existing one.alias Order alias Order.{Id, Note} %{id: Id --- &quot;o123556&quot;} |&gt; Order.new!() |&gt; Order.put!(:note, Note --- &quot;Deliver on Tue&quot;)The dialyzer can check if properly tagged values are passed as parameters to these functions. The new!/1 function can be overridden to make data validations.After the module compilation, the Domo library checks if all tags that are used with the ---/2 operator are defined and appropriately aliased. The following options can be passed with use Domo, ...Options* `undefined_tag_error_as_warning` - if set to true, prints warning instead of raising an exception for undefined tags. * `no_field` - if set to true, skips import of typedstruct/1 and field/3 macros, useful with the import of the Ecto.Schema in the same module.ReflexionEach struct or tag defines __tags__/0 function that returns a list of tags defined in the module. Additionally each tag module defines __tag__?/0 function that returns true.For example:iex.(1)&gt; defmodule Order do ....(1)&gt; use Domo ....(1)&gt; ....(1)&gt; deftag Id, for_type: String.t() ....(1)&gt; ....(1)&gt; deftag Quantity do ....(1)&gt; for_type __MODULE__.Kilograms.t() | __MODULE__.Units.t() ....(1)&gt; ....(1)&gt; deftag Kilograms, for_type: float ....(1)&gt; deftag Units, for_type: integer ....(1)&gt; end ....(1)&gt; ....(1)&gt; deftag Note, for_type: :none | String.t() ....(1)&gt; ....(1)&gt; typedstruct do ....(1)&gt; field :id, Id.t() ....(1)&gt; field :quantity, Quantity.t() ....(1)&gt; field :note, Note.t(), default: Note --- :none ....(1)&gt; end ....(1)&gt; end {:module, Order, &lt;&lt;70, 79, 82, 49, 0, 0, 17, 156, 66, 69, 65, 77, 65, 116, 85, 56, 0, 0, 1, 131, 0, 0, 0, 41, 12, 69, 108, 105, 120, 105, 114, 46, 79, 114, 100, 101, 114, 8, 95, 95, 105, 110, 102, 111, 95, 95, 7, ...&gt;&gt;, [put!: 3, put!: 3, put!: 3, put!: 3, put!: 3]} iex.(2)&gt; Order.__tags__ [Order.Id, Order.Quantity, Order.Note] iex.(3)&gt; Order.Id.__tag__? truePipelandTo add a tag or a tag chain to a value in a pipe use tag/2 macro and to remove use untag!/2 macro appropriately.For instance:import Domo alias Order.Id identifier |&gt; untag!(Id) |&gt; String.graphemes() |&gt; Enum.intersperse(&quot;_&quot;) |&gt; Enum.join() |&gt; tag(Id)LimitationsOnly a map is expected as the fields argument for the new!/1 function because it's impossible to define a typespec for a keyword list with a set of required key-value pairs.It's still possible to construct a struct with a new/1 function with fields of the wrong shape at runtime overlooking the dialyzer warnings.","ref":"Domo.html","title":"Domo","type":"module"},{"doc":"Defines a tagged tuple inline.The operator is right-associative. It adds a tag or a chain of tags to a value.Examplesiex&gt; import Domo ...&gt; Tag --- 12 {Tag, 12}","ref":"Domo.html#---/2","title":"Domo.---/2","type":"macro"},{"doc":"Defines a tag for a type.The macro generates a module with a given name that is an atom and can be used as a tag in a tagged tuple.The generated module defines a @type t(), a tagged tuple where the first element is a module's name, and the second element is a type of the value.It can be called in one-line and block forms.Examples# Define a tag as a submodule named ExperienceYears # Colleague.ExperienceYears.t() is {Colleague.ExperienceYears, integer} iex&gt; defmodule Colleague do ...&gt; import Domo ...&gt; ...&gt; deftag ExperienceYears, for_type: integer ...&gt; ...&gt; @type seniority() :: ExperienceYears.t() ...&gt; endIn the block form, you can specify the for_type/1 macro. The macro is required and should be passed within the do: block. It's possible to add other tags into the current one.Examplesiex&gt; import Domo ...&gt; deftag Email do ...&gt; for_type :none | Unverified.t() | Verified.t() ...&gt; ...&gt; deftag Unverified, for_type: String.t() ...&gt; deftag Verified, for_type: String.t() ...&gt; end ...&gt; ...&gt; Email --- Email.Unverified --- &quot;some@server.com&quot; {DomoTest.Email, {DomoTest.Email.Unverified, &quot;some@server.com&quot;}}","ref":"Domo.html#deftag/2","title":"Domo.deftag/2","type":"macro"},{"doc":"Defines a field in a typed struct.Example# A field named :example of type String.t() field :example, String.t() field :title, String.t(), default: &quot;Hello world!&quot;Optionsdefault - sets the default value for the field","ref":"Domo.html#field/2","title":"Domo.field/2","type":"macro"},{"doc":"","ref":"Domo.html#field/3","title":"Domo.field/3","type":"macro"},{"doc":"Defines a tagged tuple type t().Exampledeftag Title do # Define a tagged tuple type spec @type t :: {__MODULE__, String.t()} for_type String.t() end","ref":"Domo.html#for_type/1","title":"Domo.for_type/1","type":"macro"},{"doc":"Returns tagged tuple by joining a tag chain with a value.The macro supports up to 6 links in the tag chain.Exampleiex&gt; import Domo ...&gt; tag(2.5, SomeTag) {SomeTag, 2.5} iex&gt; import Domo ...&gt; tag(7, A --- Tag --- Chain) {A, {Tag, {Chain, 7}}}","ref":"Domo.html#tag/2","title":"Domo.tag/2","type":"macro"},{"doc":"Defines a struct with all keys enforced, a new!/1, and a put!/2 functions into it.The macro defines a struct by passing an [enforced: true] option, and the do block to the typed_struct function of the same-named library. See the TypedStruct documentation for syntax details.The default implementation of the new!/1 constructor function looks like the following:def new!(map), do: struct!(__MODULE__, map)The function can be overridden.The put!/2 function should be used to update the existing structure. Its definition looks like the following:def put!(%__MODULE__{} = s, field, value)Both new!/1 and put!/2 have type specs defined from the struct fields. Meaning, that the dialyzer can indicate contract break when values with wrong tags are used to construct or modify the structure.Unfortunately, the dialyzer can't analyze wrongly typed values mixed with correctly typed in the keyword list. Because of that, the function taking a list of fields and values to be updated is impossible to define.Examplesiex&gt; defmodule Person do ...&gt; use Domo ...&gt; ...&gt; @typedoc &quot;A person&quot; ...&gt; typedstruct do ...&gt; field :name, String.t() ...&gt; end ...&gt; end ...&gt; ...&gt; p = Person.new!(%{name: &quot;Sarah Connor&quot;}) ...&gt; Person.put!(p, :name, &quot;Connor&quot;)All defined fields are enforced automatically. We can specify an optional field with an atom and override new!/1 to verify values before construction.iex&gt; defmodule Hero do ...&gt; use Domo ...&gt; ...&gt; @typedoc &quot;A hero&quot; ...&gt; typedstruct do ...&gt; field :name, String.t() ...&gt; field :optional_kid, :none | String.t(), default: :none ...&gt; end ...&gt; ...&gt; def new!(name) when is_binary(name), do: super(%{name: name}) ...&gt; ...&gt; def new!(%{optional_kid: kid} = map) when kid in [&quot;John Connor&quot;, :none], ...&gt; do: super(map) ...&gt; end ...&gt; ...&gt; Hero.new!(&quot;Sarah Connor&quot;) ...&gt; Hero.new!(%{name: &quot;Sarah Connor&quot;, optional_kid: &quot;John Connor&quot;})","ref":"Domo.html#typedstruct/1","title":"Domo.typedstruct/1","type":"macro"},{"doc":"Returns a value from a tagged tuple when a tag chain matches.Raises ArgumentError exception if the passed tag chain is not one that is in the tagged tuple. Supports up to 6 links in the tag chain.Examplesiex&gt; import Domo ...&gt; value = A --- Tag --- Chain --- 2 ...&gt; untag!(value, A --- Tag --- Chain) 2 # When the value is a tagged tuple with the different tag, we get an ArgumentError exception value = Other --- Stuff --- 2 untag!(value, A --- Tag --- Chain) # ArgumentError, Tag chain {A, {Tag, Chain}} doesn&#39;t match one in the tagged tuple {Other, {Stuff, 2}}.","ref":"Domo.html#untag!/2","title":"Domo.untag!/2","type":"macro"},{"doc":"Module to validate tags used in the module","ref":"Domo.CompilationChecks.html","title":"Domo.CompilationChecks","type":"module"},{"doc":"","ref":"Domo.CompilationChecks.html#warn_and_raise_undefined_tags/2","title":"Domo.CompilationChecks.warn_and_raise_undefined_tags/2","type":"function"}]